import keyserver.MathOps;
import tickets.Tickets;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.Vector;


public class KeyServer 
{	 
//	private static String ticketsFile = "./tickets.txt";
//	private static int ticketsToRead = 1000;
//	private static int ticketsBitlength = 64;
//	private static int pBitlength = ticketsBitlength - 2;
	
	private static String ticketsFile;
	private static String refreshmentsFile;
	private static String timeResultsFile;
	private static int numberOfRefreshments;
	private static int numberOfTickets;
	private static int ticketsBitlength;
	private static int mBitlength;
	private static Vector<BigInteger> tickets;
	private static BigInteger L;
	private static long accumulatedTime;
	private static long averageTime;
	private static boolean refreshmentsToFile; 

	/*
	private static void example1()
	{
		Vector<BigInteger> tickets;
		BigInteger memberTicket;
		BigInteger[] extendedResult;
		BigInteger[] mpq;
		BigInteger[] kdelta;
		BigInteger[] duv;
		BigInteger u;
		BigInteger v;
		BigInteger k;
		BigInteger delta;
		BigInteger L;
		BigInteger m, p, q;
		BigInteger g;
		BigInteger r;
		
		long startTime;
		long endTime;
		
		try
		{
			//KEYSERVER SIDE
			System.out.println("KEYSERVER SIDE");
			System.out.println();
			
			//1. generate a tickets file
			//System.out.println(Tickets.storeNewTicketRange("./tickets.txt", 1000, 64));
			
			//2. read tickets from file
			System.out.print("Reading tickets from "+ticketsFile+" ...");
			tickets = Tickets.getRangeFromFile(ticketsFile, 0, ticketsToRead-1);
			System.out.println(tickets.size() +" "+ticketsBitlength+"-bit tickets read");
			
			//3. Calculate the productory L
			System.out.println();
			System.out.print("Calculating productory L...");
			L = MathOps.productory(tickets);
			System.out.println(" bitlength "+L.bitLength());
			
			startTime = System.nanoTime();
			
			//4. Calculate a refreshment of r
			//4.1. Calculate m, p, q
			System.out.println();
			System.out.println("Calculating m, p, q...");
			mpq = MathOps.findMPQ(pBitlength);
			m = mpq[0];
			p = mpq[1];
			q = mpq[2];
			System.out.println("m: "+m+" bitlength "+m.bitLength());
			System.out.println("p: "+p+" bitlength "+p.bitLength());
			System.out.println("q: "+q+" bitlength "+q.bitLength());
			
			//4.2. Calculate k and delta
			System.out.println();
			System.out.println("Calculating k, delta...");
			kdelta = MathOps.findKDelta(p, p.bitLength());
			k = kdelta[0];
			delta = kdelta[1];
			System.out.println("k: "+k+" bitlength "+k.bitLength());
			System.out.println("delta: "+delta+" bitlength "+delta.bitLength());
			
			//4.3. Calculate g
			System.out.println();
			System.out.println("Calculating g...");
			g = MathOps.findGenerator(m, p, q);
			System.out.println("g: "+g+" bitlength "+g.bitLength());
			
			//4.4. Calculate r
			System.out.println();
			System.out.println("Calculating r...");
			r = MathOps.calculateSecret(g, k, m);
			System.out.println("r: "+r+" bitlength "+r.bitLength());
			
			//4.5 Calculate u
			System.out.println();
			System.out.println("Calculating u...");
			duv = MathOps.extendedEuclides(delta, L);
			u = duv[1];
			v = duv[2];
			System.out.println("u bitlength "+u.bitLength());
			System.out.println("v bitlength "+v.bitLength());
			
			endTime = System.nanoTime();
			System.out.println();
			System.out.println("Processing time at KeyServer: "+((endTime-startTime)/1.0E09)+" secs");
			
			//MEMBER SIDE
			System.out.println();
			System.out.println();
			System.out.println("MEMBER SIDE");
			System.out.println();
			
			//1. read a ticket from the tickets file
			System.out.print("Reading ticket from "+ticketsFile+"...");
			memberTicket = Tickets.getTicketFromFile(ticketsFile, 1);
			System.out.println(" "+memberTicket.bitLength()+" bits read");
			
			startTime = System.nanoTime();
			
			//2. Retrieve r
			//Note that the member knows g, m and u
			
			//2.1. Retrieve delta
			System.out.println("Retrieving delta...");
			BigInteger deltaMember = u.modInverse(memberTicket);
			System.out.println("delta: "+deltaMember);
			
			//2.2. Retrieve r
			System.out.println("Retrieving r...");
			BigInteger rMember = g.modPow(deltaMember, m);
			System.out.println("r: "+rMember);
			
			endTime = System.nanoTime();
			System.out.println();
			System.out.println("Processing time at member: "+((endTime-startTime)/1.0E09)+" secs");
		}
		catch(Exception e)
		{
			System.out.println("Error: "+e.getMessage());
		}

	}*/

	public static String[] readConfigFile(String configFile)
	{
		BufferedReader in;
		String line;
		String[] config = new String[4];
		int i = 0;
		
		try
		{
			in = new BufferedReader(new FileReader(configFile));
			
			while((line=in.readLine())!=null)
			{
//				line = in.readLine();
				if(!line.startsWith("#"))
				{
					config[i] = line;
					i++;
				}
			}
//			}while(line!=null);
		
			if(config[3].compareTo("")==0)
				return null;
			
			return config;
		}
		catch(Exception e)
		{
			System.out.println("Error while reading config file: either the file does not exist or its sintax is incorrect.");
			return null;
		}
	}
	
	/*
	 * Performs a single full refreshment (all parameters are renewed)
	 */
	public static boolean fullRefreshment(BigInteger L, PrintWriter output, boolean verbose)
	{
		BigInteger[] mpq;
		BigInteger[] kdelta;
		BigInteger[] duv;
		BigInteger g;
		BigInteger r;
		long startTime;
		long endTime;
		long startVerboseTime = 0;
		long endVerboseTime = 0;
		
		try
		{
			mBitlength = ticketsBitlength - 2;
	
			startTime = System.nanoTime();
			//1. Calculate m, p, q
			if(verbose) startVerboseTime = System.nanoTime(); 
			mpq = MathOps.findMPQ(mBitlength);
			if(verbose)
			{
				endVerboseTime = System.nanoTime();
				System.out.println("findMPQ: "+(endVerboseTime-startVerboseTime)/1.0E09);
			}
			//2. Calculate k, delta
			if(verbose) startVerboseTime = System.nanoTime();
			kdelta = MathOps.findKDelta(mpq[1], mpq[1].bitLength());
			if(verbose)
			{
				endVerboseTime = System.nanoTime();
				System.out.println("findKDelta: "+(endVerboseTime-startVerboseTime)/1.0E09);
			}
			//3. Calculate g
			if(verbose) startVerboseTime = System.nanoTime();
			g = MathOps.findGenerator(mpq[0], mpq[1], mpq[2]);
			if(verbose)
			{
				endVerboseTime = System.nanoTime();
				System.out.println("findGenerator: "+(endVerboseTime-startVerboseTime)/1.0E09);
			}
			//4. Calculate r
			if(verbose) startVerboseTime = System.nanoTime();
			r = MathOps.calculateSecret(g, kdelta[0], mpq[0]);
			if(verbose)
			{
				endVerboseTime = System.nanoTime();
				System.out.println("calculateSecret: "+(endVerboseTime-startVerboseTime)/1.0E09);
			}
			//5. Calculate d, u, v
			if(verbose) startVerboseTime = System.nanoTime();
			duv = MathOps.extendedEuclides(kdelta[1], L);
			if(verbose)
			{
				endVerboseTime = System.nanoTime();
				System.out.println("extendedEuclides: "+(endVerboseTime-startVerboseTime)/1.0E09);
			}
			endTime = System.nanoTime();
			
			accumulatedTime += (endTime-startTime);
			
			//print output
			if(output!=null)
			{
				output.println(g+" "+mpq[0]+" "+duv[1]);	//the message {g, m, u}
				output.println(r);	//r
			}
				
			return true;
		}
		catch(Exception e)
		{
			return false;
		}
	}
	
	/*
	 * Usage: java KeyServer tickets_file refreshments_output_file number_refreshments
	 */
	public static void main(String[] args)
	{
		PrintWriter p = null;
		PrintWriter pTime;
		String[] config;
		boolean refreshmentSuccessfull;
		boolean verbose;
		long startVerboseTime = 0;
		long endVerboseTime = 0;
		int i;
		int per10;
		
		if(args.length<1)
		{
			System.out.println("Usage: java KeyServer config_file -options");
			return;
		}
		
		try
		{
			//read configuration
			config = readConfigFile(args[0]);
			ticketsFile = config[0];
			refreshmentsFile = config[1];
			numberOfRefreshments = Integer.parseInt(config[2]);
			timeResultsFile = config[3];
			
			//print refreshments to a file?
			if(args.length>1 && args[1].compareTo("-norefreshfile")==0)
				refreshmentsToFile = false;
			else
				refreshmentsToFile = true;
			
			//show fine grain times
			if(args.length > 2 && args[2].compareTo("-verbose")==0)
				verbose = true;
			else
				verbose = false;
			
			if(refreshmentsToFile)
				p = new PrintWriter(new BufferedWriter(new FileWriter(refreshmentsFile)));
			pTime = new PrintWriter(new BufferedWriter(new FileWriter(timeResultsFile,true)));	//append, create if does not exist
			
			//esto habr� que llev�rseo a otro m�todo en el futuro, para la versi�n multiproceso 
			 
			//read tickets
			tickets = Tickets.getAllTicketsFromFile(ticketsFile);
			numberOfTickets = tickets.size();
			ticketsBitlength = tickets.get(0).bitLength();
			//calculate L
			if(verbose) startVerboseTime = System.nanoTime();
			L = MathOps.productory(tickets);
			if(verbose)
			{
				endVerboseTime = System.nanoTime();
				System.out.println("productory: "+(endVerboseTime-startVerboseTime)/1.0E09);
			}
			//start refreshments
			accumulatedTime = 0;
			for(i = 1; i <= numberOfRefreshments; i++)
			{
				if(!fullRefreshment(L, p, verbose))
				{
					System.out.println("There was an error in refreshment "+i+". Simulation stopped.");
					break;
				}
				if(i%10==0)
					System.out.println(i+" simulations done.");
			}
			
			System.out.println("End of simulation.");
			System.out.println("Refreshments made: "+(i-1)+".");
			System.out.println("Accumulated time: "+(accumulatedTime/1.0E09)+" secs.");
			averageTime = accumulatedTime/(i-1);
			System.out.println("Avg. refreshment time: "+(averageTime/1.0E09)+" secs.");
			if(refreshmentsToFile)
				System.out.println("See "+refreshmentsFile+" for details.");
			
			//close output file
			if(p!=null)
				p.close();
			
			pTime.println(tickets.size()+"-"+ticketsBitlength+": "+(averageTime/1.0E09));
			pTime.println();
			pTime.close();
		}
		catch(Exception e)
		{
			System.out.println("There was an error: "+e.getMessage());
		}
	}
	
	

    
    
    
    
}
