package keyserver;

import java.math.BigInteger;
import java.security.SecureRandom;

public class Pruebas 
{	
	private static boolean passesMillerRabin(BigInteger number, int iterations) 
    {
    	BigInteger ONE = BigInteger.ONE;
    	BigInteger TWO = BigInteger.valueOf(2);
    	
    	// Find a and m such that m is odd and this == 1 + 2**a * m
            BigInteger thisMinusOne = number.subtract(ONE);
    	BigInteger m = thisMinusOne;
    	int a = m.getLowestSetBit();
    	m = m.shiftRight(a);

    	// Do the tests
            SecureRandom rnd = new SecureRandom();
    	for (int i=0; i<iterations; i++) {
    	    // Generate a uniform random on (1, this)
    	    BigInteger b;
    	    do {
    		b = new BigInteger(number.bitLength(), rnd);
    	    } while (b.compareTo(ONE) <= 0 || b.compareTo(number) >= 0);

    	    int j = 0;
    	    BigInteger z = b.modPow(m, number);
    	    while(!((j==0 && z.equals(ONE)) || z.equals(thisMinusOne))) {
    		if (j>0 && z.equals(ONE) || ++j==a)
    		    return false;
    		z = z.modPow(TWO, number);
    	    }
    	}
    	return true;
    }

//	   public static BigInteger findSmallPrime(int bitlength)
//	    {
//		   BigInteger TWO = BigInteger.valueOf(2);
//	    	BigInteger candidate;
//	    	SecureRandom sr = new SecureRandom();
//	    	
//	    	//generate p
//	    	//pick a random candidate and make sure it is odd
//	    	do
//	    	{
//	    		candidate = new BigInteger(bitlength, sr);
//	    	}while(candidate.bitLength()!=bitlength);
//	    	//assure it is odd
//	    	if(!candidate.testBit(0))
//	    		candidate = candidate.add(BigInteger.ONE);
//	    	
//	    	for(;;)
//	    	{
//	    		if(passesMillerRabin(candidate,5))
//	    			break;	//success
//	    		//try the next odd
//	    		candidate = candidate.add(TWO);
//	    	}
//	    	System.out.println(candidate.bitLength());
//	    	return candidate;
//	    }
	
//	public static BigInteger findPrime(int bitlength)
//	{
//		BigInteger r;
//		BigInteger s;
//		BigInteger p;
//		BigInteger p0;
//		BigInteger rs;
//		BigInteger k = BigInteger.ZERO;
//		SecureRandom sr = new SecureRandom();
//		
//		//get r and s
//		r = findSmallPrime(bitlength/2);
//		s =findSmallPrime(bitlength/2);
//		//calculate p0
//		rs = r.multiply(s);
//		p0 = s.modPow(r.subtract(BigInteger.ONE), rs).subtract(r.modPow(s.subtract(BigInteger.ONE), rs)).mod(rs);
//		if(!p0.testBit(0))	//if even
//			p0 = p0.add(rs);
//		do
//		{
//			k = k.add(BigInteger.ONE);
//			p = p0.add(BigInteger.valueOf(2).multiply(k).multiply(rs));
//		}while(!passesMillerRabin(p,5));
//		
//		return p;
//	}
	
//	public static BigInteger findPrimeRecursive(int bitlength)
//	{
//		BigInteger r;
//		BigInteger s;
//		BigInteger p;
//		BigInteger p0;
//		BigInteger rs;
//		BigInteger k = BigInteger.ZERO;
//		BigInteger amount;
//		BigInteger addend = BigInteger.ZERO;
//		
//		if(bitlength <= 64)
//			return BigInteger.probablePrime(bitlength, new SecureRandom());
//		
//		//get r and s
//		r = findPrimeRecursive(bitlength/2);
//		s =findPrimeRecursive(bitlength/2);
//		//calculate p0
//		rs = r.multiply(s);
//		p0 = s.modPow(r.subtract(BigInteger.ONE), rs).subtract(r.modPow(s.subtract(BigInteger.ONE), rs)).mod(rs);
//		if(!p0.testBit(0))	//if even
//			p0 = p0.add(rs);
////		amount = BigInteger.valueOf(2).multiply(rs);
////		p = p0;
//		do
//		{
//						k = k.add(BigInteger.ONE);
////			addend = addend.add(amount);
//			p = p0.add(BigInteger.valueOf(2).multiply(k).multiply(rs));
////			p = p0.add(addend);
//			
////			p = p.add(addend);
//			
//		}while(!passesMillerRabin(p,5));
//		
//		System.out.println("Found "+p.bitLength()+" bits prime");
//		return p;
//	}
	
//	//Returns p, strong prime of "bitlength", and r, s primes (possible strong) such that
//	// r | (p-1) and s | (p+1)
//	//Method can be found in J. Gordon. "Strong primes are easy to find". I've added recursivity.
//	public static BigInteger findPrimeRecursiveOld(int bitlength)
//	{
//		BigInteger r;
//		BigInteger s;
//		BigInteger p;
//		BigInteger p0;
//		BigInteger rs;
//		BigInteger amount;
//		
//		if(bitlength <= 64)
//			return BigInteger.probablePrime(bitlength, new SecureRandom());
//		
//		//get r and s
//		r = findPrimeRecursiveOld(bitlength/2);
//		s =findPrimeRecursiveOld(bitlength/2);
//		//calculate p0
//		rs = r.multiply(s);
//		p0 = s.modPow(r.subtract(BigInteger.ONE), rs).subtract(r.modPow(s.subtract(BigInteger.ONE), rs)).mod(rs);
//		if(!p0.testBit(0))	//if even
//			p0 = p0.add(rs);
//		amount = rs.shiftLeft(1);	//multiply by 2
//		p = p0;
//		do
//		{
//			p = p.add(amount);	
//		}while(!passesMillerRabin(p,5));
//		
//		System.out.println("Found "+p.bitLength()+" bits prime");
//		return p;
//	}

	//Returns p, strong prime of "bitlength", and r, s primes (possible strong) such that
	// r | (p-1) and s | (p+1)
	//Method can be found in J. Gordon. "Strong primes are easy to find". I've added recursivity.
	public static BigInteger[] findPrimeRecursive(int bitlength)
	{
		BigInteger r;
		BigInteger s;
		BigInteger p;
		BigInteger p0;
		BigInteger rs;
		BigInteger amount;
		
		if(bitlength < 64)
			return new BigInteger[] {BigInteger.probablePrime(bitlength, new SecureRandom()), BigInteger.ZERO, BigInteger.ZERO};
		
		//get r and s
		r = findPrimeRecursive(bitlength/2)[0];
		s =findPrimeRecursive(bitlength/2)[0];
		//calculate p0
		rs = r.multiply(s);
		p0 = s.modPow(r.subtract(BigInteger.ONE), rs).subtract(r.modPow(s.subtract(BigInteger.ONE), rs)).mod(rs);
		if(!p0.testBit(0))	//if even
			p0 = p0.add(rs);
		amount = rs.shiftLeft(1);	//multiply by 2
		p = p0;
		do
		{
			p = p.add(amount);	
		}while(!passesMillerRabin(p,5));
		
		System.out.println("Found "+p.bitLength()+" bits prime");
		System.out.println("\t p0: "+p0.bitLength());
		System.out.println("\t amount: "+amount.bitLength());
		System.out.println("\t r: "+r.bitLength());
		System.out.println("\t s: "+s.bitLength());
		return new BigInteger[] {p, r, s};
	}


	public static BigInteger[] findPrimeRecursive2(int bitlength, int bottomTopBitlength)
	{
		BigInteger r;
		BigInteger s;
		BigInteger p;
		BigInteger p0;
		BigInteger rs;
		BigInteger amount;
		
		if(bitlength <= bottomTopBitlength)
			return new BigInteger[] {BigInteger.probablePrime(bitlength, new SecureRandom()), BigInteger.ZERO, BigInteger.ZERO};
		
		//get r and s
		r = findPrimeRecursive2((bitlength/2)-10, bottomTopBitlength)[0];
		s =findPrimeRecursive2((bitlength/2)-10, bottomTopBitlength)[0];
		//calculate p0
		rs = r.multiply(s);
		p0 = s.modPow(r.subtract(BigInteger.ONE), rs).subtract(r.modPow(s.subtract(BigInteger.ONE), rs)).mod(rs);
		if(!p0.testBit(0))	//if even
			p0 = p0.add(rs);
		amount = rs.shiftLeft(1);	//multiply by 2
		p = p0;
		
		for(;;)
		{
			p = p.add(amount);	
			if(p.bitLength() < bitlength)
				continue;
			if(passesMillerRabin(p,5))
				break;
		}
		
		System.out.println("Found "+p.bitLength()+" bits prime");
		System.out.println("\t p0: "+p0.bitLength());
		System.out.println("\t amount: "+amount.bitLength());
		System.out.println("\t r: "+r.bitLength());
		System.out.println("\t s: "+s.bitLength());
		return new BigInteger[] {p, r, s};
	}
	
	public static void main(String args[])
	{
		long startTime, endTime;
		
		startTime = System.nanoTime();
		BigInteger[] prs = findPrimeRecursive2(256,128);
		endTime = System.nanoTime();
		System.out.println(prs[0].bitLength()+" bits");
		System.out.println("Time: "+(endTime-startTime)/1.0E09+" secs.");
		System.out.println("Is prime: "+prs[0].isProbablePrime(10));
		
	}

}
