import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.math.BigInteger;
import java.util.StringTokenizer;

import member.MathOps;
import tickets.Tickets;

public class Member 
{
	private static long accumulatedTime;

	
	/*
	 * Command line:
	 * java Member refreshments_file tickets_file <output_file> <-verbose>
	 */
	public static void main(String args[])
	{		
		String refreshments_file = null;
		String tickets_file = null;
		String output_file = null;
		boolean verbose = false;
		BigInteger ticket;
		BufferedReader in;
		PrintStream pTime;
		String line_read = "";
		int refreshments_OK = 0;
		int refreshments_failed = 0;
		int refreshments_total = 0;
		StringTokenizer st;
		long averageTime;
		String config[];
		int numMembers;
		int ticketLength;
		String timeResultsFile;
		
		BigInteger g;
		BigInteger m;
		BigInteger u;
		BigInteger r_file;
		
		if(args.length<1)
		{
			//System.out.println("Usage: java Member refreshments_file tickets_file -options");
			System.out.println("Usage: java Member config_file -options");
			return;
		}
		
		try
		{
			//read config file
			config = readConfigFile(args[0]);
			tickets_file = config[0];
			refreshments_file = config[1];
			numMembers = Integer.parseInt(config[2]);
			ticketLength = Integer.parseInt(config[3]);
			timeResultsFile = config[4];
			//read refreshments file
//			refreshments_file = args[0];
			//read tickets file
//			tickets_file = args[1];
			//output file?
//			if(args.length>2 && args[2].compareTo("-verbose")!=0)
//					output_file = args[2];
			//verbose?
			//if(args.length==4 && args[3].compareTo("-verbose")==0 || args.length==4 &&  args[3].compareTo("-verbose")==0)
			//	verbose = true;
			
			//read a ticket from tickets_file
			ticket = readTicketFromFile(tickets_file);
	
			//get the refreshments file
			in = new BufferedReader(new FileReader(refreshments_file));
			
			accumulatedTime = 0;
			//start with refreshments
			for(;;)
			{
				//refreshments file format: g m (newline) u (newline) r
				if((line_read = in.readLine())==null) break;
				st = new StringTokenizer(line_read, " ");
				g = new BigInteger(st.nextToken());
				m = new BigInteger(st.nextToken());
				
				//if((line_read = in.readLine())==null) break;
				u = new BigInteger(st.nextToken());
				
				//skip a line
				//in.readLine();
				if((line_read = in.readLine())==null) break;
				r_file = new BigInteger(line_read);
				
				//check the result
				if(r_file.compareTo(calculateRefreshment(g, m, u, ticket))==0)
					refreshments_OK++;
				else
					refreshments_failed++;
				
				refreshments_total++;
				if(refreshments_total%10==0)
					System.out.println(refreshments_total+" refreshments checked.");
			}
			
			averageTime = accumulatedTime/refreshments_total;
			
			System.out.println("End of simulation");
			System.out.println("Refreshments calculated: "+refreshments_total);
			System.out.println("\t Refreshments OK:        "+refreshments_OK);
			System.out.println("\t Refreshments failed:    "+refreshments_failed);
			System.out.println("Accumulated time: " + (accumulatedTime/1.0E09) + " secs");
			System.out.println("Avg. refreshment time: " + (averageTime/1.0E09) + " secs");
		
			//print results to file
			pTime = new PrintStream(new FileOutputStream(timeResultsFile,true));	//append, create if does not exist
			pTime.println(numMembers + "-" + ticketLength+": "+(averageTime/1.0E09) + " secs avg.");
			pTime.println();
			pTime.close();
			
		}
		catch(Exception e)
		{
			System.out.println("There was an error: "+e.getMessage());
		}
	}
	
	/*
	 * Returns a BigInteger representing a ticket from a tickets file.
	 * 
	 * Return an IOException if the tickets file does not have the expected format.
	 */
	private static BigInteger readTicketFromFile(String tickets_file) throws IOException
	{
		BufferedReader in = new BufferedReader(new FileReader(tickets_file));
		//skip first line
		in.readLine();
		
		return new BigInteger(in.readLine());
	}
	
	/*
	 * Returns the secret value r given the public values g, m and u
	 */
	private static BigInteger calculateRefreshment(BigInteger g, BigInteger m, BigInteger u, BigInteger ticket)
	{
		long startTime, endTime;
		BigInteger delta, r;
		
		//2. Retrieve r
		//Note that the member knows g, m and u
		//2.1. Retrieve delta
		//System.out.println("Retrieving delta...");
		//BigInteger deltaMember = u.modInverse(memberTicket);
		//System.out.println("delta: "+deltaMember);
		//2.2. Retrieve r
		//System.out.println("Retrieving r...");
		//BigInteger rMember = g.modPow(deltaMember, m);
		//System.out.println("r: "+rMember);
		
		startTime = System.nanoTime();
		
		delta = MathOps.getDelta(u, ticket);
		r = MathOps.getR(g, delta, m);
	
		endTime = System.nanoTime();
		accumulatedTime += (endTime-startTime);

		return r;
	}

	/*
	 * Read execution settings
	 * Should return a String[5]: {tickets_file, refreshments_file, number_of_members, tickets_length, time_results_file}
	 */
	public static String[] readConfigFile(String configFile)
	{
		BufferedReader in;
		String line;
		String[] config = new String[5];
		int i = 0;
		
		try
		{
			in = new BufferedReader(new FileReader(configFile));
			
			while((line=in.readLine())!=null)
			{
				if(!line.startsWith("#"))
				{
					config[i] = line;
					i++;
				}
			}
			in.close();
			
			if(config[4].compareTo("")==0)
				return null;
			
			return config;
		}
		catch(Exception e)
		{
			System.out.println("Error while reading config file: either the file does not exist or its sintax is incorrect.");
			return null;
		}
	}
	
}
